/*
 * Copyright (C) 2005-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/** \file ortp_recv.c paraslash's ortp receiver */

#include <ortp/ortp.h>
#include "para.h"

#include "ortp.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "ortp_recv.cmdline.h"

#include "error.h"
#include "audiod.h"
#include "string.h"

#define CHUNK_SIZE 128 * 1024

/**
 * data specific to the ortp receiver
 *
 * \sa receiver receiver_node
 */
struct private_ortp_recv_data {
/**
 *
 *
 * whether a header was received
 *
 * A flag indicating whether the ortp receiver already received a packet which
 * contains the audio file header.
 *
 * This flag has no effect if the audio stream indicates that no extra headers
 * will be sent (mp3).  Otherwise, all data packets are dropped until the
 * header is received.
 */
int have_header;
/** the ortp session this instance of the receiver belongs to */
RtpSession *session;
/** the time the first data or header packet was received */
struct timeval start;
/** ETA of the next chunk */
struct timeval next_chunk;
/** number of consecutive bad chunks */
int c_bad;
/** the current timestamp which is passed to the receiving function of libortp */
uint32_t timestamp;
/** \a timestamp increases by this amount */
uint32_t chunk_ts;
};


static int msg_to_buf(mblk_t *mp, char *buffer, int len)
{
	mblk_t *m, *mprev;
	size_t mlen, rlen = len;

	m = mp->b_cont;
	mprev = mp;
	while (m != NULL) {
		mlen = m->b_wptr - m->b_rptr;
		if (mlen <= rlen) {
			mblk_t *consumed = m;
			memcpy(buffer, m->b_rptr, mlen);
			/* go to next mblk_t */
			mprev->b_cont = m->b_cont;
			m = m->b_cont;
			consumed->b_cont = NULL;
			freeb (consumed);
			buffer += mlen;
			rlen -= mlen;
		} else {
			memcpy (buffer, m->b_rptr, rlen);
			m->b_rptr += rlen;
			return len;
		}
	}
	return len - rlen;
}


static void ortp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_ortp_recv_data *pord = rn->private_data;
	struct timeval tmp;

	if (tv_diff(now, &pord->next_chunk, &tmp) >= 0) {
		tmp.tv_sec = 0;
		tmp.tv_usec = 1000;
	}
	if (tv_diff(&s->timeout, &tmp, NULL) > 0)
		s->timeout = tmp;
	t->ret = 1;
}

static void compute_next_chunk(unsigned chunk_time,
		struct private_ortp_recv_data *pord)
{
	struct timeval chunk_tv = {0, chunk_time};
	struct timeval tmp;

	tv_add(&chunk_tv, &pord->next_chunk, &tmp);
	pord->next_chunk = tmp;
	pord->timestamp += pord->chunk_ts;
//	PARA_DEBUG_LOG("next chunk (ts = %d) due at %lu:%lu\n",
//		pord->timestamp, pord->next_chunk.tv_sec,
//		pord->next_chunk.tv_usec);
}

static void ortp_recv_post_select(__a_unused struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_ortp_recv_data *pord = rn->private_data;
	mblk_t *mp;
	int packet_type, stream_type;
	char tmpbuf[CHUNK_SIZE + 3];
	unsigned chunk_time;
	size_t packet_size;

//	PARA_INFO_LOG("rn: %p, pord: %p, session: %p\n", rn, pord, pord->session);
	t->ret = -E_ORTP_RECV_EOF;
	if (rn->output_eof && *rn->output_eof) {
		rn->eof = 1;
		return;
	}
	t->ret = 1;
	if (pord->start.tv_sec)
		if (tv_diff(now, &pord->next_chunk, NULL) < 0)
			return;
	mp = rtp_session_recvm_with_ts(pord->session, pord->timestamp);
	if (!mp) {
		struct timeval min_delay = {0, 100};
//		PARA_INFO_LOG("nope, chunk_ts = %d, loaded: %d, bad: %d\n",
//			pord->timestamp, rn->loaded, pord->c_bad);
		pord->c_bad++;
		t->ret = -E_TOO_MANY_BAD_CHUNKS;
		if ((pord->c_bad > 5000 && pord->start.tv_sec) || pord->c_bad > 10000)
			return;
		t->ret = 1;
		tv_add(now, &min_delay, &pord->next_chunk);
		return;
	}
	/* okay, we have a chunk of data */
	if (!pord->start.tv_sec)
		pord->start = *now;
	t->ret = msg_to_buf(mp, tmpbuf, CHUNK_SIZE);
	if (t->ret < ORTP_AUDIO_HEADER_LEN) {
		rn->eof = 1;
		if (t->ret < 0)
			t->ret = -E_MSG_TO_BUF;
		else
			t->ret = -E_ORTP_RECV_EOF;
		goto err_out;
	}
	packet_size = t->ret;
	packet_type = READ_PACKET_TYPE(tmpbuf);
	stream_type = READ_STREAM_TYPE(tmpbuf);
	chunk_time = READ_CHUNK_TIME(tmpbuf);
	pord->chunk_ts = READ_CHUNK_TS(tmpbuf);
//	PARA_INFO_LOG("packet type: %d, stream type: %d, chunk time: %u,"
//		" chunk_ts: %u, loaded: %u, bad: %d\n", packet_type,
//		stream_type, chunk_time, pord->chunk_ts, rn->loaded, pord->c_bad);
	switch (packet_type) {
	unsigned header_len, payload_len;
	case ORTP_EOF:
		rn->eof = 1;
		t->ret = -E_ORTP_RECV_EOF;
		goto err_out;
	case ORTP_BOF:
		PARA_INFO_LOG("bof (%zu)\n", packet_size);
		pord->have_header = 1;
		/* fall through */
	case ORTP_DATA:
		if (!pord->have_header && stream_type)
		/* can't use the data, wait for header */
			goto success;
		t->ret = -E_OVERRUN;
		if (packet_size + rn->loaded >= CHUNK_SIZE + ORTP_AUDIO_HEADER_LEN)
			goto err_out;
		if (packet_size > ORTP_AUDIO_HEADER_LEN) {
			memcpy(rn->buf + rn->loaded, tmpbuf + ORTP_AUDIO_HEADER_LEN,
				packet_size - ORTP_AUDIO_HEADER_LEN);
			rn->loaded += packet_size - ORTP_AUDIO_HEADER_LEN;
		}
		goto success;
	case ORTP_HEADER:
		header_len = READ_HEADER_LEN(tmpbuf);
		PARA_DEBUG_LOG("header packet (%zu bytes), header len: %d\n",
			packet_size, header_len);
		if (!pord->have_header) {
			pord->have_header = 1;
			memcpy(rn->buf, tmpbuf + ORTP_AUDIO_HEADER_LEN,
				packet_size - ORTP_AUDIO_HEADER_LEN);
			rn->loaded = packet_size - ORTP_AUDIO_HEADER_LEN;
			goto success;
		}
		t->ret = -E_INVALID_HEADER;
		if (header_len + ORTP_AUDIO_HEADER_LEN > packet_size)
			goto err_out;
		payload_len = packet_size - ORTP_AUDIO_HEADER_LEN - header_len;
		t->ret = -E_OVERRUN;
		if (rn->loaded + payload_len > CHUNK_SIZE)
			goto err_out;
		if (payload_len)
			memcpy(rn->buf + rn->loaded, tmpbuf
				+ (packet_size - payload_len), payload_len);
		rn->loaded += payload_len;
		goto success;
	}
success:
	t->ret = 1;
	freemsg(mp);
	if (pord->c_bad) {
		pord->c_bad = 0;
		pord->next_chunk = *now;
	}
	compute_next_chunk(chunk_time, pord);
	return;
err_out:
	rn->eof = 1;
	freemsg(mp);
}

static void ortp_shutdown(void)
{
//	ortp_global_stats_display();
	ortp_exit();
}

static void ortp_recv_close(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord = rn->private_data;

	rtp_session_destroy(pord->session);
	free(rn->private_data);
	free(rn->buf);
}

static void *ortp_recv_parse_config(int argc, char **argv)
{
	int ret;

	struct ortp_recv_args_info *tmp =
		para_calloc(sizeof(struct ortp_recv_args_info));

	ret = ortp_recv_cmdline_parser(argc, argv, tmp)? -E_ORTP_SYNTAX : 1;
	if (ret > 0)
		return tmp;
	free(tmp);
	return NULL;
}

static int ortp_recv_open(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord;
	struct ortp_recv_args_info *c = rn->conf;

	rn->buf = para_calloc(CHUNK_SIZE);

	rn->private_data = para_calloc(sizeof(struct private_ortp_recv_data));
	pord = rn->private_data;
	pord->session = rtp_session_new(RTP_SESSION_RECVONLY);
	PARA_NOTICE_LOG("receiving from %s:%d\n", c->host_arg, c->port_arg);
	rtp_session_set_local_addr(pord->session, c->host_arg, c->port_arg);
	rtp_session_set_remote_addr(pord->session, c->host_arg, c->port_arg);
	rtp_session_set_payload_type(pord->session, PAYLOAD_AUDIO_CONTINUOUS);
	rtp_session_enable_adaptive_jitter_compensation(pord->session, TRUE);
	rtp_session_set_jitter_compensation(pord->session,
			c->jitter_compensation_arg);
	return 1;
}

/**
 * the init function of the ortp receiver
 *
 * \param r pointer to the receiver struct to initialize
 *
 * Initialize all function pointers of \a r and call libortp's ortp_init().
 */
void ortp_recv_init(struct receiver *r)
{
	r->shutdown = ortp_shutdown;
	r->open = ortp_recv_open;
	r->close = ortp_recv_close;
	r->pre_select = ortp_recv_pre_select;
	r->post_select = ortp_recv_post_select;
	r->parse_config = ortp_recv_parse_config;

	ortp_init();
}
